# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.
#

"""Hold GenBank data in a straightforward format.

Classes:
 - Record - All of the information in a GenBank record.
 - Reference - hold reference data for a record.
 - Feature - Hold the information in a Feature Table.
 - Qualifier - Qualifiers on a Feature.

"""

import Bio.GenBank


def _wrapped_genbank(information, indent, wrap_space=1, split_char=" "):
    """Write a line of GenBank info that can wrap over multiple lines (PRIVATE).

    This takes a line of information which can potentially wrap over
    multiple lines, and breaks it up with carriage returns and
    indentation so it fits properly into a GenBank record.

    Arguments:
     - information - The string holding the information we want
       wrapped in GenBank method.
     - indent - The indentation on the lines we are writing.
     - wrap_space - Whether or not to wrap only on spaces in the
       information.
     - split_char - A specific character to split the lines on. By default
       spaces are used.

    """
    info_length = Record.GB_LINE_LENGTH - indent

    if not information:
        # GenBank files use "." for missing data
        return ".\n"

    if wrap_space:
        info_parts = information.split(split_char)
    else:
        cur_pos = 0
        info_parts = []
        while cur_pos < len(information):
            info_parts.append(information[cur_pos : cur_pos + info_length])
            cur_pos += info_length

    # first get the information string split up by line
    output_parts = []
    cur_part = ""
    for info_part in info_parts:
        if len(cur_part) + 1 + len(info_part) > info_length:
            if cur_part:
                if split_char != " ":
                    cur_part += split_char
                output_parts.append(cur_part)
            cur_part = info_part
        else:
            if cur_part == "":
                cur_part = info_part
            else:
                cur_part += split_char + info_part

    # add the last bit of information to the output
    if cur_part:
        output_parts.append(cur_part)

    # now format the information string for return
    output_info = output_parts[0] + "\n"
    for output_part in output_parts[1:]:
        output_info += " " * indent + output_part + "\n"

    return output_info


def _indent_genbank(information, indent):
    """Write out information with the specified indent (PRIVATE).

    Unlike _wrapped_genbank, this function makes no attempt to wrap
    lines -- it assumes that the information already has newlines in the
    appropriate places, and will add the specified indent to the start of
    each line.
    """
    # split the info into lines based on line breaks
    info_parts = information.split("\n")

    # the first line will have no indent
    output_info = info_parts[0] + "\n"
    for info_part in info_parts[1:]:
        output_info += " " * indent + info_part + "\n"

    return output_info


class Record:
    """Hold GenBank information in a format similar to the original record.

    The Record class is meant to make data easy to get to when you are
    just interested in looking at GenBank data.

    Attributes:
     - locus - The name specified after the LOCUS keyword in the GenBank
       record. This may be the accession number, or a clone id or something else.
     - size - The size of the record.
     - residue_type - The type of residues making up the sequence in this
       record. Normally something like RNA, DNA or PROTEIN, but may be as
       esoteric as 'ss-RNA circular'.
     - data_file_division - The division this record is stored under in
       GenBank (ie. PLN -> plants; PRI -> humans, primates; BCT -> bacteria...)
     - date - The date of submission of the record, in a form like '28-JUL-1998'
     - accession - list of all accession numbers for the sequence.
     - nid - Nucleotide identifier number.
     - pid - Proteint identifier number
     - version - The accession number + version (ie. AB01234.2)
     - db_source - Information about the database the record came from
     - gi - The NCBI gi identifier for the record.
     - keywords - A list of keywords related to the record.
     - segment - If the record is one of a series, this is info about which
       segment this record is (something like '1 of 6').
     - source - The source of material where the sequence came from.
     - organism - The genus and species of the organism (ie. 'Homo sapiens')
     - taxonomy - A listing of the taxonomic classification of the organism,
       starting general and getting more specific.
     - references - A list of Reference objects.
     - comment - Text with any kind of comment about the record.
     - features - A listing of Features making up the feature table.
     - base_counts - A string with the counts of bases for the sequence.
     - origin - A string specifying info about the origin of the sequence.
     - sequence - A string with the sequence itself.
     - contig - A string of location information for a CONTIG in a RefSeq file
     - project - The genome sequencing project numbers
       (will be replaced by the dblink cross-references in 2009).
     - dblinks - The genome sequencing project number(s) and other links.
       (will replace the project information in 2009).

    """

    # constants for outputting GenBank information
    GB_LINE_LENGTH = 79
    GB_BASE_INDENT = 12
    GB_FEATURE_INDENT = 21
    GB_INTERNAL_INDENT = 2
    GB_OTHER_INTERNAL_INDENT = 3
    GB_FEATURE_INTERNAL_INDENT = 5
    GB_SEQUENCE_INDENT = 9

    BASE_FORMAT = "%-" + str(GB_BASE_INDENT) + "s"
    INTERNAL_FORMAT = (
        " " * GB_INTERNAL_INDENT + "%-" + str(GB_BASE_INDENT - GB_INTERNAL_INDENT) + "s"
    )
    OTHER_INTERNAL_FORMAT = (
        " " * GB_OTHER_INTERNAL_INDENT
        + "%-"
        + str(GB_BASE_INDENT - GB_OTHER_INTERNAL_INDENT)
        + "s"
    )

    BASE_FEATURE_FORMAT = "%-" + str(GB_FEATURE_INDENT) + "s"
    INTERNAL_FEATURE_FORMAT = (
        " " * GB_FEATURE_INTERNAL_INDENT
        + "%-"
        + str(GB_FEATURE_INDENT - GB_FEATURE_INTERNAL_INDENT)
        + "s"
    )
    SEQUENCE_FORMAT = "%" + str(GB_SEQUENCE_INDENT) + "s"

    def __init__(self):
        """Initialize."""
        self.accession = []
        self.base_counts = ""
        self.comment = ""
        self.contig = ""
        self.data_file_division = ""
        self.date = ""
        self.db_source = ""
        self.dblinks = []
        self.definition = ""
        self.features = []
        self.gi = ""
        self.keywords = []
        self.locus = ""
        self.molecule_type = ""
        self.nid = ""
        self.organism = ""
        self.origin = ""
        self.pid = ""
        self.primary = []
        self.projects = []
        self.references = []
        self.residue_type = ""
        self.segment = ""
        self.sequence = ""
        self.size = ""
        self.source = ""
        self.taxonomy = []
        self.topology = ""
        self.version = ""
        self.wgs = ""
        self.wgs_scafld = []

    def __str__(self):
        """Provide a GenBank formatted output option for a Record.

        The objective of this is to provide an easy way to read in a GenBank
        record, modify it somehow, and then output it in 'GenBank format.'
        We are striving to make this work so that a parsed Record that is
        output using this function will look exactly like the original
        record.

        Much of the output is based on format description info at:

        ftp://ncbi.nlm.nih.gov/genbank/gbrel.txt
        """
        output = self._locus_line()
        output += self._definition_line()
        output += self._accession_line()
        output += self._version_line()
        output += self._project_line()
        output += self._dblink_line()
        output += self._nid_line()
        output += self._pid_line()
        output += self._keywords_line()
        output += self._db_source_line()
        output += self._segment_line()
        output += self._source_line()
        output += self._organism_line()
        for reference in self.references:
            output += str(reference)
        output += self._comment_line()
        output += self._features_line()
        for feature in self.features:
            output += str(feature)
        output += self._base_count_line()
        output += self._origin_line()
        output += self._sequence_line()
        output += self._wgs_line()
        output += self._wgs_scafld_line()
        output += self._contig_line()
        output += "//"
        return output

    def _locus_line(self):
        """Provide the output string for the LOCUS line (PRIVATE)."""
        output = "LOCUS"
        output += " " * 7  # 6-12 spaces
        output += "%-9s" % self.locus
        output += " "  # 22 space
        output += "%7s" % self.size
        if "PROTEIN" in self.residue_type:
            output += " aa"
        else:
            output += " bp "

        # treat circular types differently, since they'll have long residue
        # types
        if "circular" in self.residue_type:
            output += "%17s" % self.residue_type
        # second case: ss-DNA types of records
        elif "-" in self.residue_type:
            output += "%7s" % self.residue_type
            output += " " * 10  # spaces for circular
        else:
            output += " " * 3  # spaces for stuff like ss-
            output += "%-4s" % self.residue_type
            output += " " * 10  # spaces for circular

        output += " " * 2
        output += "%3s" % self.data_file_division
        output += " " * 7  # spaces for 56-63
        output += "%11s" % self.date
        output += "\n"
        return output

    def _definition_line(self):
        """Provide output for the DEFINITION line (PRIVATE)."""
        output = Record.BASE_FORMAT % "DEFINITION"
        output += _wrapped_genbank(self.definition + ".", Record.GB_BASE_INDENT)
        return output

    def _accession_line(self):
        """Output for the ACCESSION line (PRIVATE)."""
        if self.accession:
            output = Record.BASE_FORMAT % "ACCESSION"

            acc_info = ""
            for accession in self.accession:
                acc_info += "%s " % accession
            # strip off an extra space at the end
            acc_info = acc_info.rstrip()
            output += _wrapped_genbank(acc_info, Record.GB_BASE_INDENT)
        else:
            output = ""

        return output

    def _version_line(self):
        """Output for the VERSION line (PRIVATE)."""
        if self.version:
            output = Record.BASE_FORMAT % "VERSION"
            output += self.version
            output += "  GI:"
            output += "%s\n" % self.gi
        else:
            output = ""
        return output

    def _project_line(self):
        output = ""
        if len(self.projects) > 0:
            output = Record.BASE_FORMAT % "PROJECT"
            output += "%s\n" % "  ".join(self.projects)
        return output

    def _dblink_line(self):
        output = ""
        if len(self.dblinks) > 0:
            output = Record.BASE_FORMAT % "DBLINK"
            dblink_info = "\n".join(self.dblinks)
            output += _wrapped_genbank(dblink_info, Record.GB_BASE_INDENT)
        return output

    def _nid_line(self):
        """Output for the NID line. Use of NID is obsolete in GenBank files (PRIVATE)."""
        if self.nid:
            output = Record.BASE_FORMAT % "NID"
            output += "%s\n" % self.nid
        else:
            output = ""
        return output

    def _pid_line(self):
        """Output for PID line. Presumedly, PID usage is also obsolete (PRIVATE)."""
        if self.pid:
            output = Record.BASE_FORMAT % "PID"
            output += "%s\n" % self.pid
        else:
            output = ""
        return output

    def _keywords_line(self):
        """Output for the KEYWORDS line (PRIVATE)."""
        output = ""
        if len(self.keywords) >= 0:
            output += Record.BASE_FORMAT % "KEYWORDS"
            keyword_info = ""
            for keyword in self.keywords:
                keyword_info += "%s; " % keyword
            # replace the ; at the end with a period
            keyword_info = keyword_info[:-2]
            keyword_info += "."

            output += _wrapped_genbank(keyword_info, Record.GB_BASE_INDENT)

        return output

    def _db_source_line(self):
        """Output for DBSOURCE line (PRIVATE)."""
        if self.db_source:
            output = Record.BASE_FORMAT % "DBSOURCE"
            output += "%s\n" % self.db_source
        else:
            output = ""
        return output

    def _segment_line(self):
        """Output for the SEGMENT line (PRIVATE)."""
        output = ""
        if self.segment:
            output += Record.BASE_FORMAT % "SEGMENT"
            output += _wrapped_genbank(self.segment, Record.GB_BASE_INDENT)
        return output

    def _source_line(self):
        """Output for SOURCE line on where the sample came from (PRIVATE)."""
        output = Record.BASE_FORMAT % "SOURCE"
        output += _wrapped_genbank(self.source, Record.GB_BASE_INDENT)
        return output

    def _organism_line(self):
        """Output for ORGANISM line with taxonomy info (PRIVATE)."""
        output = Record.INTERNAL_FORMAT % "ORGANISM"
        # Now that species names can be too long, this line can wrap (Bug 2591)
        output += _wrapped_genbank(self.organism, Record.GB_BASE_INDENT)
        output += " " * Record.GB_BASE_INDENT
        taxonomy_info = ""
        for tax in self.taxonomy:
            taxonomy_info += "%s; " % tax
        # replace the ; at the end with a period
        taxonomy_info = taxonomy_info[:-2]
        taxonomy_info += "."
        output += _wrapped_genbank(taxonomy_info, Record.GB_BASE_INDENT)

        return output

    def _comment_line(self):
        """Output for the COMMENT lines (PRIVATE)."""
        output = ""
        if self.comment:
            output += Record.BASE_FORMAT % "COMMENT"
            output += _indent_genbank(self.comment, Record.GB_BASE_INDENT)
        return output

    def _features_line(self):
        """Output for the FEATURES line (PRIVATE)."""
        output = ""
        if len(self.features) > 0:
            output += Record.BASE_FEATURE_FORMAT % "FEATURES"
            output += "Location/Qualifiers\n"
        return output

    def _base_count_line(self):
        """Output for the BASE COUNT line with base information (PRIVATE)."""
        output = ""
        if self.base_counts:
            output += Record.BASE_FORMAT % "BASE COUNT  "
            # split up the base counts into their individual parts
            count_parts = self.base_counts.split(" ")
            while "" in count_parts:
                count_parts.remove("")
            # deal with the standard case, with a normal origin line
            # like: 474 a    356 c    428 g    364 t
            if len(count_parts) % 2 == 0:
                while len(count_parts) > 0:
                    count_info = count_parts.pop(0)
                    count_type = count_parts.pop(0)

                    output += "%7s %s" % (count_info, count_type)
            # deal with ugly ORIGIN lines like:
            # 1311257 a2224835 c2190093 g1309889 t
            # by just outputting the raw information
            else:
                output += self.base_counts
            output += "\n"
        return output

    def _origin_line(self):
        """Output for the ORIGIN line (PRIVATE)."""
        output = ""
        # only output the ORIGIN line if we have a sequence
        if self.sequence:
            output += Record.BASE_FORMAT % "ORIGIN"
            if self.origin:
                output += _wrapped_genbank(self.origin, Record.GB_BASE_INDENT)
            else:
                output += "\n"
        return output

    def _sequence_line(self):
        """Output for all of the sequence (PRIVATE)."""
        output = ""
        if self.sequence:
            cur_seq_pos = 0
            while cur_seq_pos < len(self.sequence):
                output += Record.SEQUENCE_FORMAT % str(cur_seq_pos + 1)

                for section in range(6):
                    start_pos = cur_seq_pos + section * 10
                    end_pos = start_pos + 10
                    seq_section = self.sequence[start_pos:end_pos]
                    output += " %s" % seq_section.lower()

                    # stop looping if we are out of sequence
                    if end_pos > len(self.sequence):
                        break

                output += "\n"
                cur_seq_pos += 60
        return output

    def _wgs_line(self):
        output = ""
        if self.wgs:
            output += Record.BASE_FORMAT % "WGS"
            output += self.wgs
        return output

    def _wgs_scafld_line(self):
        output = ""
        if self.wgs_scafld:
            output += Record.BASE_FORMAT % "WGS_SCAFLD"
            output += self.wgs_scafld
        return output

    def _contig_line(self):
        """Output for CONTIG location information from RefSeq (PRIVATE)."""
        output = ""
        if self.contig:
            output += Record.BASE_FORMAT % "CONTIG"
            output += _wrapped_genbank(
                self.contig, Record.GB_BASE_INDENT, split_char=","
            )
        return output


class Reference:
    """Hold information from a GenBank reference.

    Attributes:
     - number - The number of the reference in the listing of references.
     - bases - The bases in the sequence the reference refers to.
     - authors - String with all of the authors.
     - consrtm - Consortium the authors belong to.
     - title - The title of the reference.
     - journal - Information about the journal where the reference appeared.
     - medline_id - The medline id for the reference.
     - pubmed_id - The pubmed_id for the reference.
     - remark - Free-form remarks about the reference.

    """

    def __init__(self):
        """Initialize."""
        self.number = ""
        self.bases = ""
        self.authors = ""
        self.consrtm = ""
        self.title = ""
        self.journal = ""
        self.medline_id = ""
        self.pubmed_id = ""
        self.remark = ""

    def __str__(self):
        """Convert the reference to a GenBank format string."""
        output = self._reference_line()
        output += self._authors_line()
        output += self._consrtm_line()
        output += self._title_line()
        output += self._journal_line()
        output += self._medline_line()
        output += self._pubmed_line()
        output += self._remark_line()

        return output

    def _reference_line(self):
        """Output for REFERENCE lines (PRIVATE)."""
        output = Record.BASE_FORMAT % "REFERENCE"
        if self.number:
            if self.bases:
                output += "%-3s" % self.number
                output += "%s" % self.bases
            else:
                output += "%s" % self.number

        output += "\n"
        return output

    def _authors_line(self):
        """Output for AUTHORS information (PRIVATE)."""
        output = ""
        if self.authors:
            output += Record.INTERNAL_FORMAT % "AUTHORS"
            output += _wrapped_genbank(self.authors, Record.GB_BASE_INDENT)
        return output

    def _consrtm_line(self):
        """Output for CONSRTM information (PRIVATE)."""
        output = ""
        if self.consrtm:
            output += Record.INTERNAL_FORMAT % "CONSRTM"
            output += _wrapped_genbank(self.consrtm, Record.GB_BASE_INDENT)
        return output

    def _title_line(self):
        """Output for TITLE information (PRIVATE)."""
        output = ""
        if self.title:
            output += Record.INTERNAL_FORMAT % "TITLE"
            output += _wrapped_genbank(self.title, Record.GB_BASE_INDENT)
        return output

    def _journal_line(self):
        """Output for JOURNAL information (PRIVATE)."""
        output = ""
        if self.journal:
            output += Record.INTERNAL_FORMAT % "JOURNAL"
            output += _wrapped_genbank(self.journal, Record.GB_BASE_INDENT)
        return output

    def _medline_line(self):
        """Output for MEDLINE information (PRIVATE)."""
        output = ""
        if self.medline_id:
            output += Record.INTERNAL_FORMAT % "MEDLINE"
            output += self.medline_id + "\n"
        return output

    def _pubmed_line(self):
        """Output for PUBMED information (PRIVATE)."""
        output = ""
        if self.pubmed_id:
            output += Record.OTHER_INTERNAL_FORMAT % "PUBMED"
            output += self.pubmed_id + "\n"
        return output

    def _remark_line(self):
        """Output for REMARK information (PRIVATE)."""
        output = ""
        if self.remark:
            output += Record.INTERNAL_FORMAT % "REMARK"
            output += _wrapped_genbank(self.remark, Record.GB_BASE_INDENT)
        return output


class Feature:
    """Hold information about a Feature in the Feature Table of GenBank record.

    Attributes:
     - key - The key name of the featue (ie. source)
     - location - The string specifying the location of the feature.
     - qualfiers - A list of Qualifier objects in the feature.

    """

    def __init__(self, key="", location=""):
        """Initialize."""
        self.key = key
        self.location = location
        self.qualifiers = []

    def __repr__(self):
        """Representation of the object for debugging or logging."""
        return "Feature(key=%r, location=%r)" % (self.key, self.location)

    def __str__(self):
        """Return feature as a GenBank format string."""
        output = Record.INTERNAL_FEATURE_FORMAT % self.key
        output += _wrapped_genbank(
            self.location, Record.GB_FEATURE_INDENT, split_char=","
        )
        for qualifier in self.qualifiers:
            output += str(qualifier)
        return output


class Qualifier:
    """Hold information about a qualifier in a GenBank feature.

    Attributes:
     - key - The key name of the qualifier (ie. /organism=)
     - value - The value of the qualifier ("Dictyostelium discoideum").

    """

    def __init__(self, key="", value=""):
        """Initialize."""
        self.key = key
        self.value = value

    def __repr__(self):
        """Representation of the object for debugging or logging."""
        return "Qualifier(key=%r, value=%r)" % (self.key, self.value)

    def __str__(self):
        """Return feature qualifier as a GenBank format string."""
        output = " " * Record.GB_FEATURE_INDENT
        # determine whether we can wrap on spaces
        space_wrap = 1
        for no_space_key in Bio.GenBank._BaseGenBankConsumer.remove_space_keys:
            if no_space_key in self.key:
                space_wrap = 0
        # return double quotes as-is, leave it to the user to escape them
        return output + _wrapped_genbank(
            self.key + self.value, Record.GB_FEATURE_INDENT, space_wrap
        )



# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.
#

"""Code for working with NMR data.

This directory currently contains contributions from:
 - Bob Bussell <rgb2003@med.cornell.edu> -- NOEtools and xpktools

"""



# Copyright 2001 by Brad Chapman.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Plots to compare information between different sources.

This file contains high level plots which are designed to be used to
compare different types of information. The most basic example is comparing
two variables in a traditional scatter plot.
"""
# reportlab
from reportlab.lib import colors
from reportlab.graphics.charts.lineplots import LinePlot
from reportlab.lib.pagesizes import letter
from reportlab.lib.units import inch

from reportlab.graphics.shapes import Drawing, String
from reportlab.graphics.charts.markers import makeEmptySquare, makeFilledSquare
from reportlab.graphics.charts.markers import makeFilledDiamond, makeSmiley
from reportlab.graphics.charts.markers import makeFilledCircle, makeEmptyCircle

from Bio.Graphics import _write


class ComparativeScatterPlot:
    """Display a scatter-type plot comparing two different kinds of info.

    Attributes;
     - display_info - a 2D list of the information we'll be outputting. Each
       top level list is a different data type, and each data point is a
       two-tuple of the coordinates of a point.

    So if you had two distributions of points, it should look like::

       display_info = [[(1, 2), (3, 4)],
                       [(5, 6), (7, 8)]]

    If everything is just one set of points, display_info can look like::

        display_info = [[(1, 2), (3, 4), (5, 6)]]

    """

    def __init__(self, output_format="pdf"):
        """Initialize."""
        # customizable attributes
        self.number_of_columns = 1
        self.page_size = letter
        self.title_size = 20

        self.output_format = output_format

        # the information we'll be writing
        self.display_info = []

        # initial colors and shapes used for drawing points
        self.color_choices = [
            colors.red,
            colors.green,
            colors.blue,
            colors.yellow,
            colors.orange,
            colors.black,
        ]
        self.shape_choices = [
            makeFilledCircle,
            makeEmptySquare,
            makeFilledDiamond,
            makeFilledSquare,
            makeEmptyCircle,
            makeSmiley,
        ]

    def draw_to_file(self, output_file, title):
        """Write the comparative plot to a file.

        Arguments:
         - output_file - The name of the file to output the information to,
           or a handle to write to.
         - title - A title to display on the graphic.

        """
        width, height = self.page_size
        cur_drawing = Drawing(width, height)

        self._draw_title(cur_drawing, title, width, height)

        start_x = inch * 0.5
        end_x = width - inch * 0.5
        end_y = height - 1.5 * inch
        start_y = 0.5 * inch
        self._draw_scatter_plot(cur_drawing, start_x, start_y, end_x, end_y)

        return _write(cur_drawing, output_file, self.output_format)

    def _draw_title(self, cur_drawing, title, width, height):
        """Add a title to the page we are outputting (PRIVATE)."""
        title_string = String(width / 2, height - inch, title)
        title_string.fontName = "Helvetica-Bold"
        title_string.fontSize = self.title_size
        title_string.textAnchor = "middle"

        cur_drawing.add(title_string)

    def _draw_scatter_plot(self, cur_drawing, x_start, y_start, x_end, y_end):
        """Draw a scatter plot on the drawing with the given coordinates (PRIVATE)."""
        scatter_plot = LinePlot()

        # set the dimensions of the scatter plot
        scatter_plot.x = x_start
        scatter_plot.y = y_start
        scatter_plot.width = abs(x_start - x_end)
        scatter_plot.height = abs(y_start - y_end)

        scatter_plot.data = self.display_info

        scatter_plot.joinedLines = 0

        # set the axes of the plot
        x_min, x_max, y_min, y_max = self._find_min_max(self.display_info)
        scatter_plot.xValueAxis.valueMin = x_min
        scatter_plot.xValueAxis.valueMax = x_max
        scatter_plot.xValueAxis.valueStep = (x_max - x_min) / 10.0

        scatter_plot.yValueAxis.valueMin = y_min
        scatter_plot.yValueAxis.valueMax = y_max
        scatter_plot.yValueAxis.valueStep = (y_max - y_min) / 10.0

        self._set_colors_and_shapes(scatter_plot, self.display_info)

        cur_drawing.add(scatter_plot)

    def _set_colors_and_shapes(self, scatter_plot, display_info):
        """Set the colors and shapes of the points displayed (PRIVATE).

        By default this just sets all of the points according to the order
        of colors and shapes defined in self.color_choices and
        self.shape_choices. The first 5 shapes and colors are unique, the
        rest of them are just set to the same color and shape (since I
        ran out of shapes!).

        You can change how this function works by either changing the
        values of the color_choices and shape_choices attributes, or
        by inheriting from this class and overriding this function.
        """
        for value_num in range(len(display_info)):
            # if we have unique colors, add them
            if (value_num + 1) < len(self.color_choices):
                scatter_plot.lines[value_num].strokeColor = self.color_choices[
                    value_num
                ]
                scatter_plot.lines[value_num].symbol = self.shape_choices[value_num]
            # otherwise just use the last number
            else:
                scatter_plot.lines[value_num].strokeColor = self.color_choices[-1]
                scatter_plot.lines[value_num].symbol = self.shape_choices[-1]

    def _find_min_max(self, info):
        """Find min and max for x and y coordinates in the given data (PRIVATE)."""
        x_min = info[0][0][0]
        x_max = info[0][0][0]
        y_min = info[0][0][1]
        y_max = info[0][0][1]

        for two_d_list in info:
            for x, y in two_d_list:
                if x > x_max:
                    x_max = x
                if x < x_min:
                    x_min = x
                if y > y_max:
                    y_max = y
                if y < y_min:
                    y_min = y

        return x_min, x_max, y_min, y_max



# Copyright 2017 by Kozo Nishida.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""Code to work with the KEGG Gene database.

Functions:
- parse - Returns an iterator giving Record objects.

Classes:
- Record - A representation of a KEGG Gene.

"""


from Bio.KEGG import _default_wrap, _wrap_kegg, _write_kegg


# Set up line wrapping rules (see Bio.KEGG._wrap_kegg)
name_wrap = [0, "", (" ", "$", 1, 1), ("-", "$", 1, 1)]
id_wrap = _default_wrap


class Record:
    """Holds info from a KEGG Gene record.

    Attributes:
     - entry       The entry identifier.
     - name        A list of the gene names.
     - definition  The definition for the gene.
     - orthology   A list of 2-tuples: (orthology id, role)
     - organism    A tuple: (organism id, organism)
     - position    The position for the gene
     - motif       A list of 2-tuples: (database, list of link ids)
     - dblinks     A list of 2-tuples: (database, list of link ids)

    """

    def __init__(self):
        """Initialize new record."""
        self.entry = ""
        self.name = []
        self.definition = ""
        self.orthology = []
        self.organism = ""
        self.position = ""
        self.motif = []
        self.dblinks = []

    def __str__(self):
        """Return a string representation of this Record."""
        return self._entry() + self._name() + self._dblinks() + "///"

    def _entry(self):
        return _write_kegg("ENTRY", [self.entry])

    def _name(self):
        return _write_kegg(
            "NAME", [_wrap_kegg(l, wrap_rule=name_wrap) for l in self.name]
        )

    def _definition(self):
        return _write_kegg("DEFINITION", [self.definition])

    def _dblinks(self):
        s = []
        for entry in self.dblinks:
            s.append(entry[0] + ": " + " ".join(entry[1]))
        return _write_kegg("DBLINKS", [_wrap_kegg(l, wrap_rule=id_wrap(9)) for l in s])


def parse(handle):
    """Parse a KEGG Gene file, returning Record objects.

    This is an iterator function, typically used in a for loop.  For
    example, using one of the example KEGG files in the Biopython
    test suite,

    >>> with open("KEGG/gene.sample") as handle:
    ...     for record in parse(handle):
    ...         print("%s %s" % (record.entry, record.name[0]))
    ...
    b1174 minE
    b1175 minD


    """
    record = Record()
    for line in handle:
        if line[:3] == "///":
            yield record
            record = Record()
            continue
        if line[:12] != "            ":
            keyword = line[:12]
        data = line[12:].strip()
        if keyword == "ENTRY       ":
            words = data.split()
            record.entry = words[0]
        elif keyword == "NAME        ":
            data = data.strip(";")
            record.name.append(data)
        elif keyword == "DEFINITION  ":
            record.definition = data
        elif keyword == "ORTHOLOGY   ":
            id, name = data.split("  ")
            orthology = (id, name)
            record.orthology.append(orthology)
        elif keyword == "ORGANISM    ":
            id, name = data.split("  ")
            organism = (id, name)
            record.organism = organism
        elif keyword == "POSITION    ":
            record.position = data
        elif keyword == "MOTIF       ":
            key, values = data.split(": ")
            values = values.split()
            row = (key, values)
            record.motif.append(row)
        elif keyword == "DBLINKS     ":
            if ":" in data:
                key, values = data.split(": ")
                values = values.split()
                row = (key, values)
                record.dblinks.append(row)
            else:
                row = record.dblinks[-1]
                key, values = row
                values.extend(data.split())
                row = key, values
                record.dblinks[-1] = row


if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest()



# Copyright 2009 by Cymon J. Cox.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Command line wrapper for the multiple alignment program Clustal W."""


import os
from Bio.Application import _Option, _Switch, AbstractCommandline


class ClustalwCommandline(AbstractCommandline):
    """Command line wrapper for clustalw (version one or two).

    http://www.clustal.org/

    Notes
    -----
    Last checked against versions: 1.83 and 2.1

    References
    ----------
    Larkin MA, Blackshields G, Brown NP, Chenna R, McGettigan PA,
    McWilliam H, Valentin F, Wallace IM, Wilm A, Lopez R, Thompson JD,
    Gibson TJ, Higgins DG. (2007). Clustal W and Clustal X version 2.0.
    Bioinformatics, 23, 2947-2948.

    Examples
    --------
    >>> from Bio.Align.Applications import ClustalwCommandline
    >>> in_file = "unaligned.fasta"
    >>> clustalw_cline = ClustalwCommandline("clustalw2", infile=in_file)
    >>> print(clustalw_cline)
    clustalw2 -infile=unaligned.fasta

    You would typically run the command line with clustalw_cline() or via
    the Python subprocess module, as described in the Biopython tutorial.

    """

    # TODO - Should we default to cmd="clustalw2" now?
    def __init__(self, cmd="clustalw", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-infile", "-INFILE", "INFILE", "infile"],
                "Input sequences.",
                filename=True,
            ),
            _Option(
                ["-profile1", "-PROFILE1", "PROFILE1", "profile1"],
                "Profiles (old alignment).",
                filename=True,
            ),
            _Option(
                ["-profile2", "-PROFILE2", "PROFILE2", "profile2"],
                "Profiles (old alignment).",
                filename=True,
            ),
            # ################# VERBS (do things) #############################
            _Switch(
                ["-options", "-OPTIONS", "OPTIONS", "options"],
                "List the command line parameters",
            ),
            _Switch(
                ["-help", "-HELP", "HELP", "help"], "Outline the command line params."
            ),
            _Switch(
                ["-check", "-CHECK", "CHECK", "check"],
                "Outline the command line params.",
            ),
            _Switch(
                ["-fullhelp", "-FULLHELP", "FULLHELP", "fullhelp"],
                "Output full help content.",
            ),
            _Switch(
                ["-align", "-ALIGN", "ALIGN", "align"], "Do full multiple alignment."
            ),
            _Switch(["-tree", "-TREE", "TREE", "tree"], "Calculate NJ tree."),
            _Switch(
                ["-pim", "-PIM", "PIM", "pim"],
                "Output percent identity matrix (while calculating the tree).",
            ),
            _Option(
                ["-bootstrap", "-BOOTSTRAP", "BOOTSTRAP", "bootstrap"],
                "Bootstrap a NJ tree (n= number of bootstraps; def. = 1000).",
                checker_function=lambda x: isinstance(x, int),
            ),
            _Switch(
                ["-convert", "-CONVERT", "CONVERT", "convert"],
                "Output the input sequences in a different file format.",
            ),
            # #################### PARAMETERS (set things) #########################
            # ***General settings:****
            # Makes no sense in biopython
            # _Option(["-interactive", "-INTERACTIVE", "INTERACTIVE", "interactive"],
            #        [],
            #        lambda x: 0, # Does not take value
            #        False,
            #        "read command line, then enter normal interactive menus",
            #        False),
            _Switch(
                ["-quicktree", "-QUICKTREE", "QUICKTREE", "quicktree"],
                "Use FAST algorithm for the alignment guide tree",
            ),
            _Option(
                ["-type", "-TYPE", "TYPE", "type"],
                "PROTEIN or DNA sequences",
                checker_function=lambda x: x in ["PROTEIN", "DNA", "protein", "dna"],
            ),
            _Switch(
                ["-negative", "-NEGATIVE", "NEGATIVE", "negative"],
                "Protein alignment with negative values in matrix",
            ),
            _Option(
                ["-outfile", "-OUTFILE", "OUTFILE", "outfile"],
                "Output sequence alignment file name",
                filename=True,
            ),
            _Option(
                ["-output", "-OUTPUT", "OUTPUT", "output"],
                "Output format: CLUSTAL(default), GCG, GDE, PHYLIP, PIR, NEXUS and FASTA",
                checker_function=lambda x: x
                in [
                    "CLUSTAL",
                    "GCG",
                    "GDE",
                    "PHYLIP",
                    "PIR",
                    "NEXUS",
                    "FASTA",
                    "clustal",
                    "gcg",
                    "gde",
                    "phylip",
                    "pir",
                    "nexus",
                    "fasta",
                ],
            ),
            _Option(
                ["-outorder", "-OUTORDER", "OUTORDER", "outorder"],
                "Output taxon order: INPUT or ALIGNED",
                checker_function=lambda x: x
                in ["INPUT", "input", "ALIGNED", "aligned"],
            ),
            _Option(
                ["-case", "-CASE", "CASE", "case"],
                "LOWER or UPPER (for GDE output only)",
                checker_function=lambda x: x in ["UPPER", "upper", "LOWER", "lower"],
            ),
            _Option(
                ["-seqnos", "-SEQNOS", "SEQNOS", "seqnos"],
                "OFF or ON (for Clustal output only)",
                checker_function=lambda x: x in ["ON", "on", "OFF", "off"],
            ),
            _Option(
                ["-seqno_range", "-SEQNO_RANGE", "SEQNO_RANGE", "seqno_range"],
                "OFF or ON (NEW- for all output formats)",
                checker_function=lambda x: x in ["ON", "on", "OFF", "off"],
            ),
            _Option(
                ["-range", "-RANGE", "RANGE", "range"],
                "Sequence range to write starting m to m+n. "
                "Input as string eg. '24,200'",
            ),
            _Option(
                ["-maxseqlen", "-MAXSEQLEN", "MAXSEQLEN", "maxseqlen"],
                "Maximum allowed input sequence length",
                checker_function=lambda x: isinstance(x, int),
            ),
            _Switch(
                ["-quiet", "-QUIET", "QUIET", "quiet"],
                "Reduce console output to minimum",
            ),
            _Option(
                ["-stats", "-STATS", "STATS", "stats"],
                "Log some alignment statistics to file",
                filename=True,
            ),
            # ***Fast Pairwise Alignments:***
            _Option(
                ["-ktuple", "-KTUPLE", "KTUPLE", "ktuple"],
                "Word size",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-topdiags", "-TOPDIAGS", "TOPDIAGS", "topdiags"],
                "Number of best diags.",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-window", "-WINDOW", "WINDOW", "window"],
                "Window around best diags.",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-pairgap", "-PAIRGAP", "PAIRGAP", "pairgap"],
                "Gap penalty",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-score", "-SCORE", "SCORE", "score"],
                "Either: PERCENT or ABSOLUTE",
                checker_function=lambda x: x
                in ["percent", "PERCENT", "absolute", "ABSOLUTE"],
            ),
            # ***Slow Pairwise Alignments:***
            _Option(
                ["-pwmatrix", "-PWMATRIX", "PWMATRIX", "pwmatrix"],
                "Protein weight matrix=BLOSUM, PAM, GONNET, ID or filename",
                checker_function=lambda x: (
                    x
                    in [
                        "BLOSUM",
                        "PAM",
                        "GONNET",
                        "ID",
                        "blosum",
                        "pam",
                        "gonnet",
                        "id",
                    ]
                    or os.path.exists(x)
                ),
                filename=True,
            ),
            _Option(
                ["-pwdnamatrix", "-PWDNAMATRIX", "PWDNAMATRIX", "pwdnamatrix"],
                "DNA weight matrix=IUB, CLUSTALW or filename",
                checker_function=lambda x: (
                    x in ["IUB", "CLUSTALW", "iub", "clustalw"] or os.path.exists(x)
                ),
                filename=True,
            ),
            _Option(
                ["-pwgapopen", "-PWGAPOPEN", "PWGAPOPEN", "pwgapopen"],
                "Gap opening penalty",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-pwgapext", "-PWGAPEXT", "PWGAPEXT", "pwgapext"],
                "Gap extension penalty",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            # ***Multiple Alignments:***
            _Option(
                ["-newtree", "-NEWTREE", "NEWTREE", "newtree"],
                "Output file name for newly created guide tree",
                filename=True,
            ),
            _Option(
                ["-usetree", "-USETREE", "USETREE", "usetree"],
                "File name of guide tree",
                checker_function=lambda x: os.path.exists,
                filename=True,
            ),
            _Option(
                ["-matrix", "-MATRIX", "MATRIX", "matrix"],
                "Protein weight matrix=BLOSUM, PAM, GONNET, ID or filename",
                checker_function=lambda x: (
                    x
                    in [
                        "BLOSUM",
                        "PAM",
                        "GONNET",
                        "ID",
                        "blosum",
                        "pam",
                        "gonnet",
                        "id",
                    ]
                    or os.path.exists(x)
                ),
                filename=True,
            ),
            _Option(
                ["-dnamatrix", "-DNAMATRIX", "DNAMATRIX", "dnamatrix"],
                "DNA weight matrix=IUB, CLUSTALW or filename",
                checker_function=lambda x: (
                    x in ["IUB", "CLUSTALW", "iub", "clustalw"] or os.path.exists(x)
                ),
                filename=True,
            ),
            _Option(
                ["-gapopen", "-GAPOPEN", "GAPOPEN", "gapopen"],
                "Gap opening penalty",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-gapext", "-GAPEXT", "GAPEXT", "gapext"],
                "Gap extension penalty",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Switch(
                ["-endgaps", "-ENDGAPS", "ENDGAPS", "endgaps"],
                "No end gap separation pen.",
            ),
            _Option(
                ["-gapdist", "-GAPDIST", "GAPDIST", "gapdist"],
                "Gap separation pen. range",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Switch(
                ["-nopgap", "-NOPGAP", "NOPGAP", "nopgap"], "Residue-specific gaps off"
            ),
            _Switch(["-nohgap", "-NOHGAP", "NOHGAP", "nohgap"], "Hydrophilic gaps off"),
            _Switch(
                ["-hgapresidues", "-HGAPRESIDUES", "HGAPRESIDUES", "hgapresidues"],
                "List hydrophilic res.",
            ),
            _Option(
                ["-maxdiv", "-MAXDIV", "MAXDIV", "maxdiv"],
                "% ident. for delay",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            # Already handled in General Settings section, but appears a second
            # time under Multiple Alignments in the help
            # _Option(["-type", "-TYPE", "TYPE", "type"],
            #        "PROTEIN or DNA",
            #        checker_function=lambda x: x in ["PROTEIN", "DNA",
            #                                         "protein", "dna"]),
            _Option(
                ["-transweight", "-TRANSWEIGHT", "TRANSWEIGHT", "transweight"],
                "Transitions weighting",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-iteration", "-ITERATION", "ITERATION", "iteration"],
                "NONE or TREE or ALIGNMENT",
                checker_function=lambda x: x
                in ["NONE", "TREE", "ALIGNMENT", "none", "tree", "alignment"],
            ),
            _Option(
                ["-numiter", "-NUMITER", "NUMITER", "numiter"],
                "maximum number of iterations to perform",
                checker_function=lambda x: isinstance(x, int),
            ),
            _Switch(
                ["-noweights", "-NOWEIGHTS", "NOWEIGHTS", "noweights"],
                "Disable sequence weighting",
            ),
            # ***Profile Alignments:***
            _Switch(
                ["-profile", "-PROFILE", "PROFILE", "profile"],
                "Merge two alignments by profile alignment",
            ),
            _Option(
                ["-newtree1", "-NEWTREE1", "NEWTREE1", "newtree1"],
                "Output file name for new guide tree of profile1",
                filename=True,
            ),
            _Option(
                ["-newtree2", "-NEWTREE2", "NEWTREE2", "newtree2"],
                "Output file for new guide tree of profile2",
                filename=True,
            ),
            _Option(
                ["-usetree1", "-USETREE1", "USETREE1", "usetree1"],
                "File name of guide tree for profile1",
                checker_function=lambda x: os.path.exists,
                filename=True,
            ),
            _Option(
                ["-usetree2", "-USETREE2", "USETREE2", "usetree2"],
                "File name of guide tree for profile2",
                checker_function=lambda x: os.path.exists,
                filename=True,
            ),
            # ***Sequence to Profile Alignments:***
            _Switch(
                ["-sequences", "-SEQUENCES", "SEQUENCES", "sequences"],
                "Sequentially add profile2 sequences to profile1 alignment",
            ),
            # These are already handled in the Multiple Alignments section,
            # but appear a second time here in the help.
            # _Option(["-newtree", "-NEWTREE", "NEWTREE", "newtree"],
            #        "File for new guide tree",
            #        filename=True),
            # _Option(["-usetree", "-USETREE", "USETREE", "usetree"],
            #        "File for old guide tree",
            #        checker_function=lambda x: os.path.exists,
            #        filename=True),
            # ***Structure Alignments:***
            _Switch(
                ["-nosecstr1", "-NOSECSTR1", "NOSECSTR1", "nosecstr1"],
                "Do not use secondary structure-gap penalty mask for profile 1",
            ),
            _Switch(
                ["-nosecstr2", "-NOSECSTR2", "NOSECSTR2", "nosecstr2"],
                "Do not use secondary structure-gap penalty mask for profile 2",
            ),
            _Option(
                ["-secstrout", "-SECSTROUT", "SECSTROUT", "secstrout"],
                "STRUCTURE or MASK or BOTH or NONE output in alignment file",
                checker_function=lambda x: x
                in [
                    "STRUCTURE",
                    "MASK",
                    "BOTH",
                    "NONE",
                    "structure",
                    "mask",
                    "both",
                    "none",
                ],
            ),
            _Option(
                ["-helixgap", "-HELIXGAP", "HELIXGAP", "helixgap"],
                "Gap penalty for helix core residues",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-strandgap", "-STRANDGAP", "STRANDGAP", "strandgap"],
                "gap penalty for strand core residues",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-loopgap", "-LOOPGAP", "LOOPGAP", "loopgap"],
                "Gap penalty for loop regions",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-terminalgap", "-TERMINALGAP", "TERMINALGAP", "terminalgap"],
                "Gap penalty for structure termini",
                checker_function=lambda x: (isinstance(x, int) or isinstance(x, float)),
            ),
            _Option(
                ["-helixendin", "-HELIXENDIN", "HELIXENDIN", "helixendin"],
                "Number of residues inside helix to be treated as terminal",
                checker_function=lambda x: isinstance(x, int),
            ),
            _Option(
                ["-helixendout", "-HELIXENDOUT", "HELIXENDOUT", "helixendout"],
                "Number of residues outside helix to be treated as terminal",
                checker_function=lambda x: isinstance(x, int),
            ),
            _Option(
                ["-strandendin", "-STRANDENDIN", "STRANDENDIN", "strandendin"],
                "Number of residues inside strand to be treated as terminal",
                checker_function=lambda x: isinstance(x, int),
            ),
            _Option(
                ["-strandendout", "-STRANDENDOUT", "STRANDENDOUT", "strandendout"],
                "Number of residues outside strand to be treated as terminal",
                checker_function=lambda x: isinstance(x, int),
            ),
            # ***Trees:***
            _Option(
                ["-outputtree", "-OUTPUTTREE", "OUTPUTTREE", "outputtree"],
                "nj OR phylip OR dist OR nexus",
                checker_function=lambda x: x
                in ["NJ", "PHYLIP", "DIST", "NEXUS", "nj", "phylip", "dist", "nexus"],
            ),
            _Option(
                ["-seed", "-SEED", "SEED", "seed"],
                "Seed number for bootstraps.",
                checker_function=lambda x: isinstance(x, int),
            ),
            _Switch(
                ["-kimura", "-KIMURA", "KIMURA", "kimura"], "Use Kimura's correction."
            ),
            _Switch(
                ["-tossgaps", "-TOSSGAPS", "TOSSGAPS", "tossgaps"],
                "Ignore positions with gaps.",
            ),
            _Option(
                ["-bootlabels", "-BOOTLABELS", "BOOTLABELS", "bootlabels"],
                "Node OR branch position of bootstrap values in tree display",
                checker_function=lambda x: x in ["NODE", "BRANCH", "node", "branch"],
            ),
            _Option(
                ["-clustering", "-CLUSTERING", "CLUSTERING", "clustering"],
                "NJ or UPGMA",
                checker_function=lambda x: x in ["NJ", "UPGMA", "nj", "upgma"],
            ),
        ]
        AbstractCommandline.__init__(self, cmd, **kwargs)


if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest()



# Copyright 1999 by Jeffrey Chang.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Code for dealing with BLAST programs and output."""



# Copyright 2017-2019 Damien Goutte-Gattat.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""Bio.SeqIO support for the "xdna" file format.

The Xdna binary format is generated by Christian Marck's DNA Strider program
and also used by Serial Cloner.
"""

from re import match
from struct import pack, unpack
import warnings

from Bio import BiopythonWarning
from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation, ExactPosition
from Bio.SeqRecord import SeqRecord
from .Interfaces import SequenceIterator, SequenceWriter


_seq_types = {
    0: None,
    1: "DNA",
    2: "DNA",
    3: "RNA",
    4: "protein",
}

_seq_topologies = {0: "linear", 1: "circular"}


def _read(handle, length):
    """Read the specified number of bytes from the given handle."""
    data = handle.read(length)
    if len(data) < length:
        raise ValueError("Cannot read %d bytes from handle" % length)
    return data


def _read_pstring(handle):
    """Read a Pascal string.

    A Pascal string comprises a single byte giving the length of the string
    followed by as many bytes.
    """
    length = unpack(">B", _read(handle, 1))[0]
    return unpack("%ds" % length, _read(handle, length))[0].decode("ASCII")


def _read_pstring_as_integer(handle):
    return int(_read_pstring(handle))


def _read_overhang(handle):
    """Read an overhang specification.

    An overhang is represented in a XDNA file as:
      - a Pascal string containing the text representation of the overhang
        length, which also indicates the nature of the overhang:
        - a length of zero means no overhang,
        - a negative length means a 3' overhang,
        - a positive length means a 5' overhang;
      - the actual overhang sequence.

    Examples:
      - 0x01 0x30: no overhang ("0", as a P-string)
      - 0x01 0x32 0x41 0x41: 5' AA overhang (P-string "2", then "AA")
      - 0x02 0x2D 0x31 0x43: 3' C overhang (P-string "-1", then "C")

    Returns a tuple (length, sequence).

    """
    length = _read_pstring_as_integer(handle)
    if length != 0:
        overhang = _read(handle, abs(length))
        return (length, overhang)
    else:
        return (None, None)


def _parse_feature_description(desc, qualifiers):
    """Parse the description field of a Xdna feature.

    The 'description' field of a feature sometimes contains several
    GenBank-like qualifiers, separated by carriage returns (CR, 0x0D).
    """
    # Split the field's value in CR-separated lines, skipping empty lines
    for line in [x for x in desc.split("\x0D") if len(x) > 0]:
        # Is it a qualifier="value" line?
        m = match('^([^=]+)="([^"]+)"?$', line)
        if m:
            # Store the qualifier as provided
            qual, value = m.groups()
            qualifiers[qual] = [value]
        elif '"' not in line:  # Reject ill-formed qualifiers
            # Store the entire line as a generic note qualifier
            qualifiers["note"] = [line]


def _read_feature(handle, record):
    """Read a single sequence feature."""
    name = _read_pstring(handle)
    desc = _read_pstring(handle)
    type = _read_pstring(handle) or "misc_feature"
    start = _read_pstring_as_integer(handle)
    end = _read_pstring_as_integer(handle)

    # Feature flags (4 bytes):
    # byte 1 is the strand (0: reverse strand, 1: forward strand);
    # byte 2 tells whether to display the feature;
    # byte 4 tells whether to draw an arrow when displaying the feature;
    # meaning of byte 3 is unknown.
    (forward, display, arrow) = unpack(">BBxB", _read(handle, 4))
    if forward:
        strand = 1
    else:
        strand = -1
        start, end = end, start

    # The last field is a Pascal string usually containing a
    # comma-separated triplet of numbers ranging from 0 to 255.
    # I suspect this represents the RGB color to use when displaying
    # the feature. Skip it as we have no need for it.
    _read_pstring(handle)

    # Assemble the feature
    # Shift start by -1 as XDNA feature coordinates are 1-based
    # while Biopython uses 0-based couting.
    location = FeatureLocation(start - 1, end, strand=strand)
    qualifiers = {}
    if name:
        qualifiers["label"] = [name]
    _parse_feature_description(desc, qualifiers)
    feature = SeqFeature(location, type=type, qualifiers=qualifiers)
    record.features.append(feature)


class XdnaIterator(SequenceIterator):
    """Parser for Xdna files."""

    def __init__(self, source):
        """Parse a Xdna file and return a SeqRecord object.

        Argument source is a file-like object in binary mode or a path to a file.

        Note that this is an "iterator" in name only since an Xdna file always
        contain a single sequence.

        """
        super().__init__(source, mode="b", fmt="Xdna")

    def parse(self, handle):
        """Start parsing the file, and return a SeqRecord generator."""
        # Parse fixed-size header and do some rudimentary checks
        #
        # The "neg_length" value is the length of the part of the sequence
        # before the nucleotide considered as the "origin" (nucleotide number 1,
        # which in DNA Strider is not always the first nucleotide).
        # Biopython's SeqRecord has no such concept of a sequence origin as far
        # as I know, so we ignore that value. SerialCloner has no such concept
        # either and always generates files with a neg_length of zero.
        header = handle.read(112)
        if not header:
            raise ValueError("Empty file.")
        if len(header) < 112:
            raise ValueError("Improper header, cannot read 112 bytes from handle")
        records = self.iterate(handle, header)
        return records

    def iterate(self, handle, header):
        """Parse the file and generate SeqRecord objects."""
        (version, seq_type, topology, length, neg_length, com_length) = unpack(
            ">BBB25xII60xI12x", header
        )
        if version != 0:
            raise ValueError("Unsupported XDNA version")
        if seq_type not in _seq_types:
            raise ValueError("Unknown sequence type")
        # Read actual sequence and comment found in all XDNA files
        sequence = _read(handle, length).decode("ASCII")
        comment = _read(handle, com_length).decode("ASCII")

        # Try to derive a name from the first "word" of the comment
        name = comment.split(" ")[0]

        # Create record object
        record = SeqRecord(Seq(sequence), description=comment, name=name, id=name)
        if _seq_types[seq_type]:
            record.annotations["molecule_type"] = _seq_types[seq_type]

        if topology in _seq_topologies:
            record.annotations["topology"] = _seq_topologies[topology]

        if len(handle.read(1)) == 1:
            # This is an XDNA file with an optional annotation section.

            # Skip the overhangs as I don't know how to represent
            # them in the SeqRecord model.
            _read_overhang(handle)  # right-side overhang
            _read_overhang(handle)  # left-side overhang

            # Read the features
            num_features = unpack(">B", _read(handle, 1))[0]
            while num_features > 0:
                _read_feature(handle, record)
                num_features -= 1

        yield record


class XdnaWriter(SequenceWriter):
    """Write files in the Xdna format."""

    def __init__(self, target):
        """Initialize an Xdna writer object.

        Arguments:
         - target - Output stream opened in binary mode, or a path to a file.

        """
        super().__init__(target, mode="wb")

    def write_file(self, records):
        """Write the specified record to a Xdna file.

        Note that the function expects a list (or iterable) of records
        as per the SequenceWriter interface, but the list should contain
        only one record as the Xdna format is a mono-record format.
        """
        records = iter(records)

        try:
            record = next(records)
        except StopIteration:
            raise ValueError("Must have one sequence") from None

        try:
            next(records)
            raise ValueError("More than one sequence found")
        except StopIteration:
            pass

        self._has_truncated_strings = False

        molecule_type = record.annotations.get("molecule_type")
        if molecule_type is None:
            seqtype = 0
        elif "DNA" in molecule_type:
            seqtype = 1
        elif "RNA" in molecule_type:
            seqtype = 3
        elif "protein" in molecule_type:
            seqtype = 4
        else:
            seqtype = 0

        if record.annotations.get("topology", "linear") == "circular":
            topology = 1
        else:
            topology = 0

        # We store the record's id and description in the comment field.
        # Make sure to avoid duplicating the id if it is already
        # contained in the description.
        if record.description.startswith(record.id):
            comment = record.description
        else:
            comment = f"{record.id} {record.description}"

        # Write header
        self.handle.write(
            pack(
                ">BBB25xII60xI11xB",
                0,  # version
                seqtype,
                topology,
                len(record),
                0,  # negative length
                len(comment),
                255,  # end of header
            )
        )

        # Actual sequence and comment
        self.handle.write(str(record.seq).encode("ASCII"))
        self.handle.write(comment.encode("ASCII"))

        self.handle.write(pack(">B", 0))  # Annotation section marker
        self._write_pstring("0")  # right-side overhang
        self._write_pstring("0")  # left-side overhand

        # Write features
        # We must skip features with fuzzy locations as they cannot be
        # represented in the Xdna format
        features = [
            f
            for f in record.features
            if type(f.location.start) == ExactPosition
            and type(f.location.end) == ExactPosition
        ]
        drop = len(record.features) - len(features)
        if drop > 0:
            warnings.warn(
                f"Dropping {drop} features with fuzzy locations", BiopythonWarning
            )

        # We also cannot store more than 255 features as the number of
        # features is stored on a single byte...
        if len(features) > 255:
            drop = len(features) - 255
            warnings.warn(
                f"Too many features, dropping the last {drop}", BiopythonWarning
            )
            features = features[:255]

        self.handle.write(pack(">B", len(features)))
        for feature in features:
            self._write_pstring(feature.qualifiers.get("label", [""])[0])

            description = ""
            for qname in feature.qualifiers:
                if qname in ("label", "translation"):
                    continue

                for val in feature.qualifiers[qname]:
                    if len(description) > 0:
                        description = description + "\x0D"
                    description = description + '%s="%s"' % (qname, val)
            self._write_pstring(description)

            self._write_pstring(feature.type)

            start = feature.location.start.position + 1  # 1-based coordinates
            end = feature.location.end.position
            strand = 1
            if feature.location.strand == -1:
                start, end = end, start
                strand = 0
            self._write_pstring(str(start))
            self._write_pstring(str(end))

            self.handle.write(pack(">BBBB", strand, 1, 0, 1))
            self._write_pstring("127,127,127")

        if self._has_truncated_strings:
            warnings.warn(
                "Some annotations were truncated to 255 characters", BiopythonWarning
            )

        return 1

    def _write_pstring(self, s):
        """Write the given string as a Pascal string."""
        if len(s) > 255:
            self._has_truncated_strings = True
            s = s[:255]
        self.handle.write(pack(">B", len(s)))
        self.handle.write(s.encode("ASCII"))



# Copyright (C) 2013 by Ben Morris (ben@bendmorris.com)
# Based on Bio.Nexus, copyright 2005-2008 by Frank Kauff & Cymon J. Cox
# and Bio.Phylo.Newick, copyright 2009 by Eric Talevich.
# All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""I/O function wrappers for the RDF/CDAO file format.

This is an RDF format that conforms to the Comparative Data Analysis Ontology (CDAO).
See: http://evolutionaryontology.org/cdao

This module requires the librdf Python bindings (http://www.librdf.org)

The CDAOIO.Parser, in addition to parsing text files, can also parse directly
from a triple store that implements the Redland storage interface; similarly,
the CDAOIO.Writer can store triples in a triple store instead of serializing
them to a file.
"""

import os

from io import StringIO

from Bio import MissingPythonDependencyError

from Bio.Phylo import CDAO
from ._cdao_owl import cdao_namespaces, resolve_uri

# import of cdao_elements from ._cdao_owl removed in Biopython 1.74


try:
    import rdflib

    rdfver = rdflib.__version__
    if rdfver[0] in ["1", "2"] or (rdfver in ["3.0.0", "3.1.0", "3.2.0"]):
        raise MissingPythonDependencyError(
            "Support for CDAO tree format requires RDFlib v3.2.1 or later."
        )
except ImportError:
    raise MissingPythonDependencyError(
        "Support for CDAO tree format requires RDFlib."
    ) from None

RDF_NAMESPACES = {
    "owl": "http://www.w3.org/2002/07/owl#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
}
RDF_NAMESPACES.update(cdao_namespaces)
# pad node ids with zeroes until they're at least this length
ZEROES = 8


def qUri(x):
    """Resolve URI for librdf."""
    return resolve_uri(x, namespaces=RDF_NAMESPACES)


def format_label(x):
    """Format label for librdf."""
    return x.replace("_", " ")


# ---------------------------------------------------------
# Public API


def parse(handle, **kwargs):
    """Iterate over the trees in a CDAO file handle.

    :returns: generator of Bio.Phylo.CDAO.Tree objects.

    """
    return Parser(handle).parse(**kwargs)


def write(trees, handle, plain=False, **kwargs):
    """Write a trees in CDAO format to the given file handle.

    :returns: number of trees written.

    """
    return Writer(trees).write(handle, plain=plain, **kwargs)


# ---------------------------------------------------------
# Input


class Parser:
    """Parse a CDAO tree given a file handle."""

    def __init__(self, handle=None):
        """Initialize CDAO tree parser."""
        self.handle = handle
        self.graph = None
        self.node_info = None
        self.children = {}
        self.rooted = False

    @classmethod
    def from_string(cls, treetext):
        """Instantiate the class from the given string."""
        handle = StringIO(treetext)
        return cls(handle)

    def parse(self, **kwargs):
        """Parse the text stream this object was initialized with."""
        self.parse_handle_to_graph(**kwargs)
        return self.parse_graph()

    def parse_handle_to_graph(
        self, rooted=False, parse_format="turtle", context=None, **kwargs
    ):
        """Parse self.handle into RDF model self.model."""
        if self.graph is None:
            self.graph = rdflib.Graph()
        graph = self.graph

        for k, v in RDF_NAMESPACES.items():
            graph.bind(k, v)

        self.rooted = rooted

        if "base_uri" in kwargs:
            base_uri = kwargs["base_uri"]
        else:
            # Windows style slashes cannot be used in an RDF URI
            base_uri = "file://" + os.path.abspath(self.handle.name).replace("\\", "/")

        graph.parse(file=self.handle, publicID=base_uri, format=parse_format)

        return self.parse_graph(graph, context=context)

    def parse_graph(self, graph=None, context=None):
        """Iterate over RDF model yielding CDAO.Tree instances."""
        if graph is None:
            graph = self.graph

        # look up branch lengths/TUs for all nodes
        self.get_node_info(graph, context=context)

        for root_node in self.tree_roots:
            clade = self.parse_children(root_node)

            yield CDAO.Tree(root=clade, rooted=self.rooted)

    def new_clade(self, node):
        """Return a CDAO.Clade object for a given named node."""
        result = self.node_info[node]

        kwargs = {}
        if "branch_length" in result:
            kwargs["branch_length"] = result["branch_length"]
        if "label" in result:
            kwargs["name"] = result["label"].replace("_", " ")
        if "confidence" in result:
            kwargs["confidence"] = result["confidence"]

        clade = CDAO.Clade(**kwargs)

        return clade

    def get_node_info(self, graph, context=None):
        """Create a dictionary containing information about all nodes in the tree."""
        self.node_info = {}
        self.obj_info = {}
        self.children = {}
        self.nodes = set()
        self.tree_roots = set()

        assignments = {
            qUri("cdao:has_Parent"): "parent",
            qUri("cdao:belongs_to_Edge_as_Child"): "edge",
            qUri("cdao:has_Annotation"): "annotation",
            qUri("cdao:has_Value"): "value",
            qUri("cdao:represents_TU"): "tu",
            qUri("rdfs:label"): "label",
            qUri("cdao:has_Support_Value"): "confidence",
        }

        for s, v, o in graph:
            # process each RDF triple in the graph sequentially

            s, v, o = str(s), str(v), str(o)

            if s not in self.obj_info:
                self.obj_info[s] = {}
            this = self.obj_info[s]

            try:
                # if the predicate is one we care about, store information for
                # later
                this[assignments[v]] = o
            except KeyError:
                pass

            if v == qUri("rdf:type"):
                if o in (qUri("cdao:AncestralNode"), qUri("cdao:TerminalNode")):
                    # this is a tree node; store it in set of all nodes
                    self.nodes.add(s)
            if v == qUri("cdao:has_Root"):
                # this is a tree; store its root in set of all tree roots
                self.tree_roots.add(o)

        for node in self.nodes:
            # for each node, look up all information needed to create a
            # CDAO.Clade
            self.node_info[node] = {}
            node_info = self.node_info[node]

            obj = self.obj_info[node]
            if "edge" in obj:
                # if this object points to an edge, we need a branch length from
                # the annotation on that edge
                edge = self.obj_info[obj["edge"]]
                if "annotation" in edge:
                    annotation = self.obj_info[edge["annotation"]]
                    if "value" in annotation:
                        node_info["branch_length"] = float(annotation["value"])

            if "tu" in obj:
                # if this object points to a TU, we need the label of that TU
                tu = self.obj_info[obj["tu"]]
                if "label" in tu:
                    node_info["label"] = tu["label"]

            if "parent" in obj:
                # store this node as a child of its parent, if it has one,
                # so that the tree can be traversed from parent to children
                parent = obj["parent"]
                if parent not in self.children:
                    self.children[parent] = []
                self.children[parent].append(node)

    def parse_children(self, node):
        """Traverse the tree to create a nested clade structure.

        Return a CDAO.Clade, and calls itself recursively for each child,
        traversing the entire tree and creating a nested structure of CDAO.Clade
        objects.
        """
        clade = self.new_clade(node)

        children = self.children[node] if node in self.children else []
        clade.clades = [self.parse_children(child_node) for child_node in children]

        return clade


# ---------------------------------------------------------
# Output


class Writer:
    """Based on the writer in Bio.Nexus.Trees (str, to_string)."""

    prefixes = RDF_NAMESPACES

    def __init__(self, trees):
        """Initialize parameters for writing a CDAO tree."""
        self.trees = trees

        self.node_counter = 0
        self.edge_counter = 0
        self.tu_counter = 0
        self.tree_counter = 0

    def write(
        self,
        handle,
        tree_uri="",
        record_complete_ancestry=False,
        rooted=False,
        **kwargs
    ):
        """Write this instance's trees to a file handle."""
        self.rooted = rooted
        self.record_complete_ancestry = record_complete_ancestry

        if tree_uri and not tree_uri.endswith("/"):
            tree_uri += "/"

        trees = self.trees

        if tree_uri:
            handle.write("@base <%s>\n" % tree_uri)
        for k, v in self.prefixes.items():
            handle.write("@prefix %s: <%s> .\n" % (k, v))

        handle.write("<%s> a owl:Ontology .\n" % self.prefixes["cdao"])

        for tree in trees:
            self.tree_counter += 1
            self.tree_uri = "tree%s"

            first_clade = tree.clade
            statements = self.process_clade(first_clade, root=tree)
            for stmt in statements:
                self.add_stmt_to_handle(handle, stmt)

    def add_stmt_to_handle(self, handle, stmt):
        """Add URI prefix to handle."""
        # apply URI prefixes
        stmt_strings = []
        for n, part in enumerate(stmt):
            if isinstance(part, rdflib.URIRef):
                node_uri = str(part)
                changed = False
                for prefix, uri in self.prefixes.items():
                    if node_uri.startswith(uri):
                        node_uri = node_uri.replace(uri, "%s:" % prefix, 1)
                        if node_uri == "rdf:type":
                            node_uri = "a"
                        changed = True
                if changed or ":" in node_uri:
                    stmt_strings.append(node_uri)
                else:
                    stmt_strings.append("<%s>" % node_uri)

            elif isinstance(part, rdflib.Literal):
                stmt_strings.append(part.n3())

            else:
                stmt_strings.append(str(part))

        handle.write("%s .\n" % " ".join(stmt_strings))

    def process_clade(self, clade, parent=None, root=False):
        """Recursively generate triples describing a tree of clades."""
        self.node_counter += 1
        clade.uri = "node%s" % str(self.node_counter).zfill(ZEROES)
        if parent:
            clade.ancestors = parent.ancestors + [parent.uri]
        else:
            clade.ancestors = []

        def nUri(s):
            # nUri = lambda s: rdflib.URIRef(s)
            return rdflib.URIRef(s)

        def pUri(s):
            # pUri = lambda s: rdflib.URIRef(qUri(s))
            return rdflib.URIRef(qUri(s))

        tree_id = nUri("")

        statements = []

        if root is not False:
            # create a cdao:RootedTree with reference to the tree root
            tree_type = (
                pUri("cdao:RootedTree") if self.rooted else pUri("cdao:UnrootedTree")
            )

            statements += [
                (tree_id, pUri("rdf:type"), tree_type),
                (tree_id, pUri("cdao:has_Root"), nUri(clade.uri)),
            ]

            try:
                tree_attributes = root.attributes
            except AttributeError:
                tree_attributes = []

            for predicate, obj in tree_attributes:
                statements.append((tree_id, predicate, obj))

        if clade.name:
            # create TU
            self.tu_counter += 1
            tu_uri = "tu%s" % str(self.tu_counter).zfill(ZEROES)

            statements += [
                (nUri(tu_uri), pUri("rdf:type"), pUri("cdao:TU")),
                (nUri(clade.uri), pUri("cdao:represents_TU"), nUri(tu_uri)),
                (
                    nUri(tu_uri),
                    pUri("rdfs:label"),
                    rdflib.Literal(format_label(clade.name)),
                ),
            ]

            try:
                tu_attributes = clade.tu_attributes
            except AttributeError:
                tu_attributes = []

            for predicate, obj in tu_attributes:
                yield (nUri(tu_uri), predicate, obj)

        # create this node
        node_type = "cdao:TerminalNode" if clade.is_terminal() else "cdao:AncestralNode"
        statements += [
            (nUri(clade.uri), pUri("rdf:type"), pUri(node_type)),
            (nUri(clade.uri), pUri("cdao:belongs_to_Tree"), tree_id),
        ]

        if parent is not None:
            # create edge from the parent node to this node
            self.edge_counter += 1
            edge_uri = "edge%s" % str(self.edge_counter).zfill(ZEROES)

            statements += [
                (nUri(edge_uri), pUri("rdf:type"), pUri("cdao:DirectedEdge")),
                (nUri(edge_uri), pUri("cdao:belongs_to_Tree"), tree_id),
                (nUri(edge_uri), pUri("cdao:has_Parent_Node"), nUri(parent.uri)),
                (nUri(edge_uri), pUri("cdao:has_Child_Node"), nUri(clade.uri)),
                (
                    nUri(clade.uri),
                    pUri("cdao:belongs_to_Edge_as_Child"),
                    nUri(edge_uri),
                ),
                (nUri(clade.uri), pUri("cdao:has_Parent"), nUri(parent.uri)),
                (
                    nUri(parent.uri),
                    pUri("cdao:belongs_to_Edge_as_Parent"),
                    nUri(edge_uri),
                ),
            ]

            try:
                confidence = clade.confidence
            except AttributeError:
                pass
            else:
                if confidence is not None:
                    confidence = rdflib.Literal(
                        confidence, datatype="http://www.w3.org/2001/XMLSchema#decimal"
                    )

                    statements += [
                        (nUri(clade.uri), pUri("cdao:has_Support_Value"), confidence)
                    ]

            if self.record_complete_ancestry and len(clade.ancestors) > 0:
                statements += [
                    (nUri(clade.uri), pUri("cdao:has_Ancestor"), nUri(ancestor))
                    for ancestor in clade.ancestors
                ]

            if clade.branch_length is not None:
                # add branch length
                edge_ann_uri = "edge_annotation%s" % str(self.edge_counter).zfill(
                    ZEROES
                )

                branch_length = rdflib.Literal(
                    clade.branch_length,
                    datatype=rdflib.URIRef("http://www.w3.org/2001/XMLSchema#decimal"),
                )
                statements += [
                    (nUri(edge_ann_uri), pUri("rdf:type"), pUri("cdao:EdgeLength")),
                    (nUri(edge_uri), pUri("cdao:has_Annotation"), nUri(edge_ann_uri)),
                    (nUri(edge_ann_uri), pUri("cdao:has_Value"), branch_length),
                ]

            try:
                edge_attributes = clade.edge_attributes
            except AttributeError:
                edge_attributes = []

            for predicate, obj in edge_attributes:
                yield (nUri(edge_uri), predicate, obj)

        yield from statements

        try:
            clade_attributes = clade.attributes
        except AttributeError:
            clade_attributes = []

        for predicate, obj in clade_attributes:
            yield (nUri(clade.uri), predicate, obj)

        if not clade.is_terminal():
            for new_clade in clade.clades:
                yield from self.process_clade(new_clade, parent=clade, root=False)



# Copyright 2013 by Iddo Friedberg idoerg@gmail.com
# Revision copyright 2013 by Peter Cock.  All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Code for dealing with assorted UniProt file formats.

This currently include parsers for the GAF, GPA and GPI formats
from UniProt-GOA as the module Bio.UniProt.GOA.

See also Bio.SwissProt and the "swiss" support in Bio.SeqIO for
the legacy plain text sequence format still used in UniProt.

See also Bio.SeqIO.SwissIO for the "uniprot-xml" support in
Bio.SeqIO.
"""



# Copyright 2003-2008 by Leighton Pritchard.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
#
# Contact:       Leighton Pritchard, The James Hutton Institute,
#                Invergowrie, Dundee, Scotland, DD2 5DA, UK
#                Leighton.Pritchard@hutton.ac.uk
"""Provides a container for information concerning the tracks to be drawn in a diagram.

It also provides the interface for defining the diagram (possibly split these
functions in later version?).

For drawing capabilities, this module uses reportlab to draw and write the
diagram:

http://www.reportlab.com

For dealing with biological information, the package expects BioPython
objects - namely SeqRecord objects containing SeqFeature objects.
"""

try:
    from reportlab.graphics import renderPM
except ImportError:
    # This is an optional part of ReportLab, so may not be installed.
    renderPM = None

from ._LinearDrawer import LinearDrawer
from ._CircularDrawer import CircularDrawer
from ._Track import Track

from Bio.Graphics import _write


def _first_defined(*args):
    """Return the first non-null argument (PRIVATE)."""
    for arg in args:
        if arg is not None:
            return arg
    return None


class Diagram:
    """Diagram container.

    Arguments:
     - name           - a string, identifier for the diagram.
     - tracks         - a list of Track objects comprising the diagram.
     - format         - a string, format of the diagram 'circular' or
       'linear', depending on the sort of diagram required.
     - pagesize       - a string, the pagesize of output describing the ISO
       size of the image, or a tuple of pixels.
     - orientation    - a string describing the required orientation of the
       final drawing ('landscape' or 'portrait').
     - x              - a float (0->1), the proportion of the page to take
       up with even X margins t the page.
     - y              - a float (0->1), the proportion of the page to take
       up with even Y margins to the page.
     - xl             - a float (0->1), the proportion of the page to take
       up with the left X margin to the page (overrides x).
     - xr             - a float (0->1), the proportion of the page to take
       up with the right X margin to the page (overrides x).
     - yt             - a float (0->1), the proportion of the page to take
       up with the top Y margin to the page (overrides y).
     - yb             - a float (0->1), the proportion of the page to take
       up with the bottom Y margin to the page (overrides y).
     - circle_core    - a float, the proportion of the available radius to
       leave empty at the center of a circular diagram (0 to 1).
     - start          - an integer, the base/aa position to start the diagram at.
     - end            - an integer, the base/aa position to end the diagram at.
     - tracklines     - a boolean, True if track guidelines are to be drawn.
     - fragments      - and integer, for a linear diagram, the number of equal
       divisions into which the sequence is divided.
     - fragment_size  - a float (0->1), the proportion of the space
       available to each fragment that should be used in drawing.
     - track_size     - a float (0->1), the proportion of the space
       available to each track that should be used in drawing with sigils.
     - circular       - a boolean, True if the genome/sequence to be drawn
       is, in reality, circular.

    """

    def __init__(
        self,
        name=None,
        format="circular",
        pagesize="A3",
        orientation="landscape",
        x=0.05,
        y=0.05,
        xl=None,
        xr=None,
        yt=None,
        yb=None,
        start=None,
        end=None,
        tracklines=False,
        fragments=10,
        fragment_size=None,
        track_size=0.75,
        circular=True,
        circle_core=0.0,
    ):
        """Initialize.

        gdd = Diagram(name=None)
        """
        self.tracks = {}  # Holds all Track objects, keyed by level
        self.name = name  # Description of the diagram
        # Diagram page setup attributes
        self.format = format
        self.pagesize = pagesize
        self.orientation = orientation
        self.x = x
        self.y = y
        self.xl = xl
        self.xr = xr
        self.yt = yt
        self.yb = yb
        self.start = start
        self.end = end
        self.tracklines = tracklines
        self.fragments = fragments
        if fragment_size is not None:
            self.fragment_size = fragment_size
        else:
            if self.fragments == 1:
                # For single fragments, default to full height
                self.fragment_size = 1
            else:
                # Otherwise keep a 10% gap between fragments
                self.fragment_size = 0.9
        self.track_size = track_size
        self.circular = circular
        self.circle_core = circle_core
        self.cross_track_links = []
        self.drawing = None

    def set_all_tracks(self, attr, value):
        """Set the passed attribute of all tracks in the set to the passed value.

        Arguments:
         - attr    - An attribute of the Track class.
         - value   - The value to set that attribute.

        set_all_tracks(self, attr, value)
        """
        for track in self.tracks.values():
            if hasattr(track, attr):
                # If the feature has the attribute set it to the passed value
                setattr(track, attr, value)

    def draw(
        self,
        format=None,
        pagesize=None,
        orientation=None,
        x=None,
        y=None,
        xl=None,
        xr=None,
        yt=None,
        yb=None,
        start=None,
        end=None,
        tracklines=None,
        fragments=None,
        fragment_size=None,
        track_size=None,
        circular=None,
        circle_core=None,
        cross_track_links=None,
    ):
        """Draw the diagram, with passed parameters overriding existing attributes.

        gdd.draw(format='circular')
        """
        # Pass the parameters to the drawer objects that will build the
        # diagrams.  At the moment, we detect overrides with an or in the
        # Instantiation arguments, but I suspect there's a neater way to do
        # this.
        if format == "linear":
            drawer = LinearDrawer(
                self,
                _first_defined(pagesize, self.pagesize),
                _first_defined(orientation, self.orientation),
                _first_defined(x, self.x),
                _first_defined(y, self.y),
                _first_defined(xl, self.xl),
                _first_defined(xr, self.xr),
                _first_defined(yt, self.yt),
                _first_defined(yb, self.yb),
                _first_defined(start, self.start),
                _first_defined(end, self.end),
                _first_defined(tracklines, self.tracklines),
                _first_defined(fragments, self.fragments),
                _first_defined(fragment_size, self.fragment_size),
                _first_defined(track_size, self.track_size),
                _first_defined(cross_track_links, self.cross_track_links),
            )
        else:
            drawer = CircularDrawer(
                self,
                _first_defined(pagesize, self.pagesize),
                _first_defined(orientation, self.orientation),
                _first_defined(x, self.x),
                _first_defined(y, self.y),
                _first_defined(xl, self.xl),
                _first_defined(xr, self.xr),
                _first_defined(yt, self.yt),
                _first_defined(yb, self.yb),
                _first_defined(start, self.start),
                _first_defined(end, self.end),
                _first_defined(tracklines, self.tracklines),
                _first_defined(track_size, self.track_size),
                _first_defined(circular, self.circular),
                _first_defined(circle_core, self.circle_core),
                _first_defined(cross_track_links, self.cross_track_links),
            )
        drawer.draw()  # Tell the drawer to complete the drawing
        self.drawing = drawer.drawing  # Get the completed drawing

    def write(self, filename="test1.ps", output="PS", dpi=72):
        """Write the drawn diagram to a specified file, in a specified format.

        Arguments:
            - filename   - a string indicating the name of the output file,
              or a handle to write to.
            - output     - a string indicating output format, one of PS, PDF,
              SVG, or provided the ReportLab renderPM module is installed, one
              of the bitmap formats JPG, BMP, GIF, PNG, TIFF or TIFF.  The
              format can be given in upper or lower case.
            - dpi        - an integer. Resolution (dots per inch) for bitmap formats.

        Returns:
            No return value.

        write(self, filename='test1.ps', output='PS', dpi=72)

        """
        return _write(self.drawing, filename, output, dpi=dpi)

    def write_to_string(self, output="PS", dpi=72):
        """Return a byte string containing the diagram in the requested format.

        Arguments:
            - output    - a string indicating output format, one of PS, PDF,
              SVG, JPG, BMP, GIF, PNG, TIFF or TIFF (as specified for the write
              method).
            - dpi       - Resolution (dots per inch) for bitmap formats.

        Returns:
            Return the completed drawing as a bytes string in a prescribed
            format.

        """
        # The ReportLab drawToString method, which this function used to call,
        # originally just used a StringIO handle with the drawToFile method.
        #
        # TODO - Rename this method to include keyword bytes?
        from io import BytesIO

        handle = BytesIO()
        self.write(handle, output, dpi)
        return handle.getvalue()

    def add_track(self, track, track_level):
        """Add a Track object to the diagram.

        It also accepts instructions to place it at a particular level on the
        diagram.

        Arguments:
            - track          - Track object to draw.
            - track_level    - an integer. The level at which the track will be
              drawn (above an arbitrary baseline).

        add_track(self, track, track_level)
        """
        if track is None:
            raise ValueError("Must specify track")
        if track_level not in self.tracks:  # No track at that level
            self.tracks[track_level] = track  # so just add it
        else:  # Already a track there, so shunt all higher tracks up one
            occupied_levels = sorted(
                self.get_levels()
            )  # Get list of occupied levels...
            occupied_levels.reverse()  # ...reverse it (highest first)
            for val in occupied_levels:
                # If track value >= that to be added
                if val >= track.track_level:
                    self.tracks[val + 1] = self.tracks[val]  # ...increment by 1
            self.tracks[track_level] = track  # And put the new track in
        self.tracks[track_level].track_level = track_level

    def new_track(self, track_level, **args):
        """Add a new Track to the diagram at a given level.

        The track is returned for further user manipulation.

        Arguments:
            - track_level   - an integer. The level at which the track will be
              drawn (above an arbitrary baseline).

        new_track(self, track_level)
        """
        newtrack = Track()
        for key in args:
            setattr(newtrack, key, args[key])
        if track_level not in self.tracks:  # No track at that level
            self.tracks[track_level] = newtrack  # so just add it
        else:  # Already a track there, so shunt all higher tracks up one
            occupied_levels = sorted(
                self.get_levels()
            )  # Get list of occupied levels...
            occupied_levels.reverse()  # ...reverse (highest first)...
            for val in occupied_levels:
                if val >= track_level:
                    # Track value >= that to be added, increment by 1
                    self.tracks[val + 1] = self.tracks[val]
            self.tracks[track_level] = newtrack  # And put the new track in
        self.tracks[track_level].track_level = track_level
        return newtrack

    def del_track(self, track_level):
        """Remove the track to be drawn at a particular level on the diagram.

        Arguments:
            - track_level   - an integer. The level of the track on the diagram
              to delete.

        del_track(self, track_level)
        """
        del self.tracks[track_level]

    def get_tracks(self):
        """Return a list of the tracks contained in the diagram."""
        return list(self.tracks.values())

    def move_track(self, from_level, to_level):
        """Move a track from one level on the diagram to another.

        Arguments:
         - from_level   - an integer. The level at which the track to be
           moved is found.
         - to_level     - an integer. The level to move the track to.

        """
        aux = self.tracks[from_level]
        del self.tracks[from_level]
        self.add_track(aux, to_level)

    def renumber_tracks(self, low=1, step=1):
        """Renumber all tracks consecutively.

        Optionally from a passed lowest number.

        Arguments:
         - low     - an integer. The track number to start from.
         - step    - an integer. The track interval for separation of
           tracks.

        """
        track = low  # Start numbering from here
        levels = self.get_levels()

        conversion = {}  # Holds new set of levels
        for level in levels:  # Starting at low...
            conversion[track] = self.tracks[level]  # Add old tracks to new set
            conversion[track].track_level = track
            track += step  # step interval
        self.tracks = conversion  # Replace old set of levels with new set

    def get_levels(self):
        """Return a sorted list of levels occupied by tracks in the diagram."""
        return sorted(self.tracks)

    def get_drawn_levels(self):
        """Return a sorted list of levels occupied by tracks.

        These tracks are not explicitly hidden.
        """
        return sorted(key for key in self.tracks if not self.tracks[key].hide)

    def range(self):
        """Return lowest and highest base numbers from track features.

        Returned type is a tuple.
        """
        lows, highs = [], []
        for track in self.tracks.values():  # Get ranges for each track
            low, high = track.range()
            lows.append(low)
            highs.append(high)
        return min(lows), max(highs)  # Return extremes from all tracks

    def __getitem__(self, key):
        """Return the track contained at the level of the passed key."""
        return self.tracks[key]

    def __str__(self):
        """Return a formatted string describing the diagram."""
        outstr = ["\n<%s: %s>" % (self.__class__, self.name)]
        outstr.append("%d tracks" % len(self.tracks))
        for level in self.get_levels():
            outstr.append("Track %d: %s\n" % (level, self.tracks[level]))
        outstr = "\n".join(outstr)
        return outstr



# Copyright 2001 by Gavin E. Crooks.  All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Handle the SCOP DEScription file.

The file format is described in the scop
"release notes.":http://scop.berkeley.edu/release-notes-1.55.html
The latest DES file can be found
"elsewhere at SCOP.":http://scop.mrc-lmb.cam.ac.uk/scop/parse/

"Release 1.55":http://scop.berkeley.edu/parse/des.cla.scop.txt_1.55 (July 2001)
"""


class Record:
    """Holds information for one node in the SCOP hierarchy.

    Attributes:
     - sunid - SCOP unique identifiers
     - nodetype - One of 'cl' (class), 'cf' (fold), 'sf' (superfamily),
       'fa' (family), 'dm' (protein), 'sp' (species), 'px' (domain).
       Additional node types may be added.
     - sccs - SCOP concise classification strings. e.g. b.1.2.1
     - name - The SCOP ID (sid) for domains (e.g. d1anu1), currently empty for other node types
     - description - e.g. "All beta proteins","Fibronectin type III",

    """

    def __init__(self, line=None):
        """Initialize the class."""
        self.sunid = ""
        self.nodetype = ""
        self.sccs = ""
        self.name = ""
        self.description = ""
        if line:
            self._process(line)

    def _process(self, line):
        """Parse DES records (PRIVATE).

        Records consist of 5 tab deliminated fields,
        sunid, node type, sccs, node name, node description.
        """
        # For example ::
        #
        # 21953   px      b.1.2.1 d1dan.1 1dan T:,U:91-106
        # 48724   cl      b       -       All beta proteins
        # 48725   cf      b.1     -       Immunoglobulin-like beta-sandwich
        # 49265   sf      b.1.2   -       Fibronectin type III
        # 49266   fa      b.1.2.1 -       Fibronectin type III

        line = line.rstrip()  # no trailing whitespace
        columns = line.split("\t")  # separate the tab-delineated cols
        if len(columns) != 5:
            raise ValueError("I don't understand the format of %s" % line)

        sunid, self.nodetype, self.sccs, self.name, self.description = columns
        if self.name == "-":
            self.name = ""
        self.sunid = int(sunid)

    def __str__(self):
        """Represent the SCOP description record as a tab-separated string."""
        s = []
        s.append(self.sunid)
        s.append(self.nodetype)
        s.append(self.sccs)
        if self.name:
            s.append(self.name)
        else:
            s.append("-")
        s.append(self.description)
        return "\t".join(map(str, s)) + "\n"


def parse(handle):
    """Iterate over a DES file as a Des record for each line.

    Arguments:
     - handle - file-like object

    """
    for line in handle:
        if line.startswith("#"):
            continue
        yield Record(line)
